;; Generated package description from parser-generator.el  -*- no-byte-compile: t -*-
(define-package "parser-generator" "0.2.8.0.20251225.164758" "Parser Generator library" '((emacs "26")) :commit "0180a672911591b86ae25fd67e99bc9b91147fe6" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://github.com/cjohansson/emacs-parser-generator")
